@@ -9,8 +9,8 @@ from typing import Any, Callable
 from libyang.data import DNode
 
 from _sysrepo import ffi, lib
-from .errors import SysrepoError, SysrepoNotFoundError, check_call
-from .util import c2str, is_async_func, xpath_split
+from .errors import SysrepoError, check_call
+from .util import c2str, is_async_func
 
 
 LOG = logging.getLogger(__name__)
@@ -32,6 +32,7 @@ class Subscription:
         private_data: Any = None,
         asyncio_register: bool = False,
         strict: bool = False,
+        include_implicit_defaults: bool = True,
     ):
         """
         :arg callback:
@@ -45,6 +46,9 @@ class Subscription:
             If True, reject data with no schema definition from rpc output parameters
             and operational data callbacks. Otherwise, ignore unknown data and log a
             warning message.
+        :arg include_implicit_defaults:
+            If True, include implicit default nodes into Change objects passed to module
+            change callbacks and into input parameters passed to RPC/action callbacks.
         """
         if is_async_func(callback) and not asyncio_register:
             raise ValueError(
@@ -54,6 +58,7 @@ class Subscription:
         self.private_data = private_data
         self.asyncio_register = asyncio_register
         self.strict = strict
+        self.include_implicit_defaults = include_implicit_defaults
         if asyncio_register:
             self.loop = asyncio.get_event_loop()
         else:
@@ -208,17 +213,16 @@ def module_change_callback(session, module, xpath, event, req_id, priv):
             if task_id not in subscription.tasks:
                 # ATTENTION: the implicit session passed as argument will be
                 # freed when this function returns. The async callback must NOT
-                # keep a reference on it as it will be invalid. Config and
-                # changes must be gathered and converted to python objects now.
-                try:
-                    config = session.get_data(
-                        root_xpath, include_implicit_defaults=True
+                # keep a reference on it as it will be invalid. Changes must be
+                # gathered now.
+                changes = list(
+                    session.get_changes(
+                        root_xpath + "//.",
+                        include_implicit_defaults=subscription.include_implicit_defaults,
                     )
-                except SysrepoNotFoundError:
-                    config = {}
-                changes = list(session.get_changes(root_xpath + "//."))
+                )
                 task = subscription.loop.create_task(
-                    callback(event_name, req_id, config, changes, private_data)
+                    callback(event_name, req_id, changes, private_data)
                 )
                 task.add_done_callback(
                     functools.partial(subscription.task_done, task_id, event_name)
@@ -239,12 +243,13 @@ def module_change_callback(session, module, xpath, event, req_id, priv):
             task.result()  # raise error if any
 
         else:
-            try:
-                config = session.get_data(root_xpath, include_implicit_defaults=True)
-            except SysrepoNotFoundError:
-                config = {}
-            changes = list(session.get_changes(root_xpath + "//."))
-            callback(event_name, req_id, config, changes, private_data)
+            changes = list(
+                session.get_changes(
+                    root_xpath + "//.",
+                    include_implicit_defaults=subscription.include_implicit_defaults,
+                )
+            )
+            callback(event_name, req_id, changes, private_data)
 
         return lib.SR_ERR_OK
 
@@ -409,22 +414,29 @@ def rpc_callback(session, xpath, input_node, event, req_id, output_node, priv):
         from .session import SysrepoSession  # circular import
 
         session = SysrepoSession(session, True)
-        ly_ctx = session.get_ly_ctx()
-        xpath = c2str(xpath)
-        input_dict = DNode.new(ly_ctx, input_node).print_dict(
-            include_implicit_defaults=True
-        )
         subscription = ffi.from_handle(priv)
         callback = subscription.callback
         private_data = subscription.private_data
         event_name = EVENT_NAMES[event]
+        ly_ctx = session.get_ly_ctx()
+        rpc_input = DNode.new(ly_ctx, input_node)
+        xpath = rpc_input.path()
+        # strip all parents, only preserve the input tree
+        input_dict = next(
+            iter(
+                rpc_input.print_dict(
+                    include_implicit_defaults=subscription.include_implicit_defaults,
+                    absolute=False,
+                ).values()
+            )
+        )
 
         if is_async_func(callback):
             task_id = (event, req_id)
 
             if task_id not in subscription.tasks:
                 task = subscription.loop.create_task(
-                    callback(input_dict, event_name, private_data)
+                    callback(xpath, input_dict, event_name, private_data)
                 )
                 task.add_done_callback(
                     functools.partial(subscription.task_done, task_id, event_name)
@@ -441,7 +453,7 @@ def rpc_callback(session, xpath, input_node, event, req_id, output_node, priv):
             output_dict = task.result()
 
         else:
-            output_dict = callback(input_dict, event_name, private_data)
+            output_dict = callback(xpath, input_dict, event_name, private_data)
 
         if event != lib.SR_EV_RPC:
             # May happen when there are multiple callback registered for the
@@ -452,15 +464,7 @@ def rpc_callback(session, xpath, input_node, event, req_id, output_node, priv):
 
         if isinstance(output_dict, dict):
             # update output_node with contents of output_dict
-            # strip first level of output_dict with the rpc name
-            prefix, name, _ = next(xpath_split(xpath))
-            name_prefix = "%s:%s" % (prefix, name)
-            if name in output_dict:
-                output_dict = output_dict[name]
-            elif name_prefix in output_dict:
-                output_dict = output_dict[name_prefix]
-            dnode = DNode.new(ly_ctx, output_node)
-            dnode.merge_data_dict(
+            DNode.new(ly_ctx, output_node).merge_data_dict(
                 output_dict, rpcreply=True, strict=subscription.strict, validate=False
             )
         elif output_dict is not None:
